#! /usr/bin/perl -wT

package Twici ;

use strict ;

use Exporter ;
our @ISA = qw(Exporter) ;
our @EXPORT = qw(handleView handleEdit) ;

use BSD::Resource ;
BEGIN
{
	setrlimit( RLIMIT_CPU, 10, 15 ) ; # 10:15 sec
	setrlimit( RLIMIT_VMEM, 52428800, 104857600 ) ; # 50:100 MB
}

### NOTES

# throughout, subroutines are called in the modern style, with parentheses and without
# ampersands (so they look like they do in real languages)

# constants are written inline where possible

# todo: use pod
# todo: in forms, make order of method + action consistent; action first

### CONFIGURATION

use constant WIKI_ROOT => "/srv/www/urchin.earth.li/users/twic/wiki/" ;
use constant WIKI_SRV => "http://urchin.earth.li/" ;
use constant WIKI_BASE => WIKI_SRV . "cgi-bin/twic/xwiki/" ;
use constant KEY_PAGES => ("FrontPage", "RecentChanges", "MessageOfTheDay") ;

### CONSTANTS

use constant WIKI_EXT => ".twici" ;
use constant BADURL_PAGE => "BadURL" ;
use constant BADPIC_PAGE => "BadPicture" ;
use constant URL_RE => "([a-zA-Z][0-9a-zA-Z+\\-\\.]*):([^\\s>]+)" ;

### HANDLERS

sub handleView
{
	($ENV{"REQUEST_METHOD"} eq "GET") || die( "script must be accessed by GET" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = parse( $queryString ) ;
	my $page = validateWikiName( $query{"page"} ) ;
	my $fh = openPage( $page, "<" ) ;
	renderPage( $page, $fh ) ;
	# todo: emit a last-modified header
}

sub handleEdit
{
	($ENV{"REQUEST_METHOD"} eq "GET") || die( "script must be accessed by GET" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = parse( $queryString ) ;
	my $page = validateWikiName( $query{"page"} ) ;
	htmlHeader( "Edit $page" ) ;
	my $action = updateUrl( $page ) ;
	print <<EOF ;
<form action="$action" method="POST">
<textarea name="text" rows="20" cols="80">
EOF
	if (pageExists( $page ))
	{
		my $fh = openPage( $page, "<" ) ;
		while (my $line = <$fh>) # todo: would io be more efficient if this was in block/slurp mode?
		{
			print $line ;
		}
		close( $fh ) ;
	}
	print <<EOF ;
</textarea>
<br>
<input type="submit">
<input type="reset">
</form>
EOF
	htmlFooter() ;
}

# todo: it would be nice and neat if view, edit and update all used blockmode. if it worked and was efficient, it should be done.

sub handleUpdate
{
	($ENV{"REQUEST_METHOD"} eq "POST") || die( "script must be accessed by POST" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = parse( $queryString ) ;
	my $page = validateWikiName( $query{"page"} ) ;
	my $contentType = $ENV{"CONTENT_TYPE"} ;
	(defined( $contentType ) && (lc( $contentType ) eq "application/x-www-form-urlencoded")) || die( "undefined or incorrect content type" ) ;
	# todo: for god's sake use CGI.pm or CGI::Lite for this bit!
	# todo: support text/twici as content-type, and just dump it to $text
	my $in ;
	my $contentLength = $ENV{"CONTENT_LENGTH"} ;
	if ($contentLength)
	{
		read( STDIN, $in, $contentLength ) ;
	}
	else # todo: would this be more efficient (and would it work) in block mode?
	{
		my $linesep = $/ ;
		$/ = undef ;
		$in = <STDIN> ;
		$/ = $linesep ;
	}
	my %form = parse( $in ) ;
	my $text = $form{"text"} ;
	defined( $text ) || die( "text not defined in form: $in" ) ;
	$text = normaliseLines( $text ) ;
	my $fh = openPage( $page, ">") ;
	print $fh $text || die( "could not write to page file: $page ($!)" ) ;
	close( $fh ) ;
	htmlHeader( "Edited $page" ) ;
	my $viewUrl = viewUrl( $page ) ;
	my $editUrl = editUrl( $page ) ;
	print <<EOF ;
<p>You have edited <a href="$viewUrl">$page</a>; you should view it to
check that your changes are okay. Alternatively, you may <a
href="$editUrl">edit</a> it again if you like.</p>
EOF
	htmlFooter() ;
}

sub handleChanges
{
	($ENV{"REQUEST_METHOD"} eq "GET") || die( "script must be accessed by GET" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = parse( $queryString ) ;
	my $days = $query{"days"} ;
	my $num = $query{"num"} ;
	htmlHeader( "Recent Changes" ) ;
	if (defined( $days ) || defined( $num ))
	{
		my $msg = "These are the " + ($num ? "last $num" : "") + " changes" + ($days ? " in the last $days days" : "") + "." ;
		print <<EOF ;
<p>$msg</p>
EOF
		my @changes = changes( $days, $num ) ;
		print "<ul>\n" ;
		foreach my $page (@changes)
		{
			my $url = viewUrl( $page ) ;
			my $mtime = pageMtime( $page ) ;
			my $mtimestr = gmtime( $mtime ) ;
			print "<li><a href=\"$url\">$page</a> ($mtimestr GMT)</li>\n" ;
		}
		print "</ul>\n" ;
	}
	else
	{
		my $action = changesUrl() ;
		print <<EOF ;
<p>This form can be used to query the wiki for a list of pages which have changed recently.</p>
<form action="$action" method="GET">
<table>
<tr><th>Days since the change</th><td><input type="text" name="days"></td></tr>
<tr><th>Number of changes since the change</th><td><input type="text" name="num"></td></tr>
</table>
<input type="submit">
<input type="reset">
</form>
<p>In both fields, blank or 0 means that the criterion is not applied.</p>
EOF
	}
	htmlFooter() ;
}

sub handleViews
{
	($ENV{"REQUEST_METHOD"} eq "GET") || die( "script must be accessed by GET" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = parse( $queryString ) ;
	my $days = $query{"days"} ;
	my $num = $query{"num"} ;
	htmlHeader( "Recent Views" ) ;
	if (defined( $days ) || defined( $num ))
	{
		my $msg = "These are the " + ($num ? "last $num" : "") + " views" + ($days ? " in the last $days days" : "") + "." ;
		print <<EOF ;
<p>$msg</p>
EOF
		my @views = views( $days, $num ) ;
		print "<ul>\n" ;
		foreach my $page (@views)
		{
			my $url = viewUrl( $page ) ;
			my $atime = pageAtime( $page ) ;
			my $atimestr = gmtime( $atime ) ;
			print "<li><a href=\"$url\">$page</a> ($atimestr GMT)</li>\n" ;
		}
		print "</ul>\n" ;
	}
	else
	{
		my $action = viewsUrl() ;
		print <<EOF ;
<p>This form can be used to query the wiki for a list of pages which have been viewed recently.</p>
<form action="$action" method="GET">
<table>
<tr><th>Days since the view</th><td><input type="text" name="days"></td></tr>
<tr><th>Number of views since the view</th><td><input type="text" name="num"></td></tr>
</table>
<input type="submit">
<input type="reset">
</form>
<p>In both fields, blank or 0 means that the criterion is not applied.</p>
EOF
	}
	htmlFooter() ;
}

sub handleBacklinks
{
	($ENV{"REQUEST_METHOD"} eq "GET") || die( "script must be accessed by GET" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = parse( $queryString ) ;
	my $page = validateWikiName( $query{"page"} ) ;
	my @backlinks = backlinks( $page ) ;
	my $pageUrl = viewUrl( $page ) ;
	htmlHeader( "Backlinks to $page" ) ;
	print <<EOF ;
<ul>
EOF
	foreach my $backlink (@backlinks)
	{
		my $backlinkUrl = viewUrl( $backlink ) ;
		print "<li><a href=\"$backlinkUrl\">$backlink</a></li>\n" ;
	}
	print <<EOF ;
</ul>
EOF
	htmlFooter() ;
}

sub handleSource
{
	($ENV{"REQUEST_METHOD"} eq "GET") || die( "script must be accessed by GET" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = parse( $queryString ) ;
	my $page = validateWikiName( $query{"page"} ) ;
	my $pageFile = pageFile( $page ) ;
	open( PAGE, "< " . $pageFile ) || die( "could not open page file" ) ;
	print "Content-type: text/plain\n" ;
	print "\n" ;
	while (my $line = <PAGE>) # todo: use block mode? see comments above ...
	{
		print $line ;
	}
}

### RENDERING

sub renderPage
{
	my $page = shift @_ ;
	my $fh = shift @_ ;
	htmlHeader( $page ) ;
	my $oldSlash = $/ ;
	$/ = "" ; # paragraph mode
	while (my $block = <$fh>)
	{
		renderBlock( $block ) ;
	}
	close( $fh ) ;
	$/ = $oldSlash ;
	htmlFooter( $page ) ;
}

sub renderBlock
{
	my $block = shift @_ ;
	$block =~ /^(.*?)\s*$/s ;
	$block = $1 ;
	my $prefix = substr( $block, 0, 1 ) ;
	if ($prefix eq "-")
	{
		renderList( $block ) ;
	}
	elsif ($prefix eq "=")
	{
		renderHeading( $block ) ;
	}
	elsif ($prefix eq "[")
	{
		renderPicture( $block ) ;
	}
	else
	{
		renderParagraph( $block ) ;
	}
}

sub renderParagraph
{
	my $para = shift @_ ;
	print "<p>" ;
	renderText( $para ) ;
	print "</p>\n" ;
}

sub renderHeading
{
	my $hdg = shift @_ ;
	$hdg =~ /^(=+)\s*(.*)$/s || die( "malformed heading: $hdg" ) ;
	my $prefix = $1 ;
	my $lev = length( $prefix ) + 1 ;
	$hdg = $2 ;
	print "<h$lev>" ;
	renderText( $hdg ) ;
	print "</h$lev>\n" ;
}

sub renderPicture
{
	my $pic = shift @_ ;
	$pic =~ /^\[(.*?)\s*\]?$/s || die( "malformed picture: $pic" ) ;
	$pic = $1 ;
	$pic =~ /^(\S*)\s*(.*)$/s || die( "malformed picture: $pic" ) ;
	my $type ;
	my $body ;
	if ($2)
	{
		$type = $1 ;
		$body = $2 ;
	}
	else
	{
		$body = $1 ;
	}
	my $html = pictureHtml( $type, $body ) ;
	print "<center>$html</center>\n" ;
}

sub pictureHtml
{
	my $type = shift ;
	my $body = shift ;
	$type || ($type = "img") ;
	$type = lc( $type ) ;
	if ($type eq "img")
	{
		$body =~ /^(\S+)\s*(.*)$/ || die ( "malformed img body: $body" ) ;
		my $url = $1 ;
		my $alt = $2 ;
		my $altStr = $alt ? " alt=\"$alt\"" : "" ;
		return "<p><img src=\"$url\"$altStr></p>" ;
	}
	else
	{
		my $url = viewUrl( BADPIC_PAGE ) ;
		return "<pre>[<a href=\"$url\">@{[BADPIC_PAGE]}</a>]</pre>" ;
	}
}

sub renderList
{
	my $list = shift @_ ;
	print "<ul>\n" ;
	my @lines = split( "\n", $list ) ;
	my $item = "" ;
	foreach my $line (@lines)
	{
		if (substr( $line, 0, 1 ) eq "-")
		{
			if ($item ne "")
			{
				renderListItem( $item ) ;
			}
			$item = substr( $line, 1 ) ; # todo: trim leading whitespace (not really necessary)
		}
		else
		{
			$item .= "\n" . $line ;
		}
	}
	if ($item ne "")
	{
		renderListItem( $item ) ;
	}
	print "</ul>\n" ;
}

sub renderListItem
{
	my $item = shift @_ ;
	print "<li>" ;
	renderText( $item ) ;
	print "</li>\n" ;
}

# this is V2.0 of the text-level renderer
# it doesn't handle unwrapped URLs
sub renderText
{
	my $text = shift @_ ;
	$text =~ tr/ \t\n/ /s ;
	my @tokens = atomise( $text ) ;
	if ($tokens[0] eq " ")
	{
		shift @tokens ;
	}
	if ($tokens[-1] eq " ")
	{
		pop @tokens ;
	}
	my $em = 0 ;
	my $strong = 0 ;
	foreach my $tok (@tokens)
	{
		if ($tok eq "_")
		{
			if (!$em)
			{
				print "<em>" ;
				$em = 1 ;
			}
			else
			{
				print "</em>" ;
				$em = 0 ;
			}
		}
		elsif ($tok eq "*")
		{
			if (!$strong)
			{
				print "<strong>" ;
				$strong = 1 ;
			}
			else
			{
				print "</strong>" ;
				$strong = 0 ;
			}
		}
		elsif (isWikiName( $tok ))
		{
                        if (pageExists( $tok ))
                        {
                                my $url = viewUrl( $tok ) ;
                                print "<a href=\"$url\">$tok</a>" ;
                        }
                        else
                        {
                                my $url = editUrl( $tok ) ;
                                print "$tok<a href=\"$url\">?</a>" ;
                        }
		}
		elsif (isUrl( $tok ))
		{
			my $url = generaliseUrl( $tok ) ;
			print "<a href=\"$url\">$tok</a>" ;
		}
		else # plain word
		{
			print escape( $tok ) ;
		}
	}
	if ($em)
	{
		print "</em>" ;
	}
	if ($strong)
	{
		print "</strong>" ;
	}
}

# breaks text up into a string of tokens.
# recognises strings of word chars and
# (almost) arbitrary strings delimited by angle brackets as single tokens,
# and calls everything else a delimiter

sub atomise
{
	my $str = shift @_ ;
	my @tokens = () ;
	while (length( $str ) > 0)
	{
		if ($str =~ /^@{[URL_RE]}/) # evilous!
		{
			my $atom = "$1:$2" ;
			push @tokens, $atom ;
			$str = substr( $str, length( $atom )) ;
		}
		elsif ($str =~ /^([A-Za-z0-9]+)/)
		{
			my $atom = $1 ;
			push @tokens, $atom ;
			$str = substr( $str, length( $atom )) ;
		}
		else
		{
			push @tokens, substr( $str, 0, 1 ) ;
			$str = substr( $str, 1 ) ;
		}
	}
	return @tokens ;
}

sub generaliseUrl
{
	my $url = shift @_ ;
	(my ($scheme, $path) = isUrl( $url )) || return viewUrl( BADURL_PAGE ) ;
	$scheme = lc( $scheme ) ;
	if ($scheme eq "isbn")
	{
		return "http://search.barnesandnoble.com/booksearch/isbninquiry.asp?ISBN=" . $path ;
	}
	elsif ($scheme eq "isfdb")
	{
		$path =~ /([A-Za-z]+)\/(.+)/ || return viewUrl( BADURL_PAGE ) ;
		my $class = lc( $1 ) ;
		my $entity = $2 ;
		if ($class eq "author")
		{
			return "http://www.sfsite.com/isfdb-bin/exact_author.cgi?" . $entity ;
		}
		elsif ($class eq "work")
		{
			return "http://www.sfsite.com/isfdb-bin/pwork.cgi?" . $entity ;
		}
		else
		{
			return viewUrl( BADURL_PAGE ) ;
		}
	}
	elsif ($scheme eq "imdb")
	{
		$path =~ /([A-Za-z]+)\/(.+)/ || return viewUrl( BADURL_PAGE ) ;
		my $class = lc( $1 ) ;
		my $entity = $2 ;
		if ($class eq "title")
		{
			$entity =~ /[0-9]+/ || return viewUrl( BADURL_PAGE ) ;
			return "http://imdb.com/Title?" . $entity ;
		}
		if ($class eq "name")
		{
			# todo: should try to validate entity
			return "http://imdb.com/Name?" . $entity ;
		}
		else
		{
			return viewUrl( BADURL_PAGE ) ;
		}
	}
	elsif ($scheme eq "lj")
	{
		# the LJ sub canonical_username says that a username is a sequence of 1-15 \w chars or hyphens
		if ($path =~ /^([\w\-]{1,15})\/([0-9]+)$/)
		{
			my $user = $1 ;
			my $item = $2 ;
			return "http://www.livejournal.com/talkread.bml?journal=$user&itemid=$item" ;
		}
		elsif ($path =~ /^[\w\-]{1,15}$/)
		{
			return "http://www.livejournal.com/users/$path" ;
		}
		else
		{
			return viewUrl( BADURL_PAGE ) ;
		}
	}
	elsif ($scheme eq "groggs")
	{
		# todo: validate ID
		return "gopher://gopher.publication.org.uk:70/1/rgtp/index%3Fgroggs/" . $path ;
	}
	elsif ($scheme eq "google")
	{
		return "http://www.google.co.uk/search?q=" . $path ;
	}
	elsif ($scheme eq "dmoz")
	{
		$path =~ /^\/?(.*?)\/?$/ ;
		$path = $1 ;
		return "http://dmoz.org/" . $path ;
	}
	else
	{
		return $url ;
	}
}

### HTML GENERATION

sub htmlHeader
{
	my $title = shift @_ ;
	print <<EOF ;
Content-type: text/html

<!doctype HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">
<html>
<head>
<title>$title</title>
</head>
<body>
<h1>$title</h1>
<hr>
EOF
}

sub htmlFooter
{
	my $page = shift @_ ;
	print <<EOF ;
<hr>
<p>
EOF
	if ($page)
	{
		my $editUrl = editUrl( $page ) ;
		my $backlinksUrl = backlinksUrl( $page ) ;
		my $sourceUrl = sourceUrl( $page ) ;
		my $validatorUrl = validatorUrl( $page ) ;
		my $mtimestr = gmtime( pageMtime( $page ) ) ;
		print <<EOF ;
| <a href="$editUrl">Edit</a>
| <a href="$backlinksUrl">Backlinks</a>
| <a href="$sourceUrl">Source</a>
| <a href="$validatorUrl">Validate</a> 
|
| Last Modified $mtimestr GMT
|
EOF
	}
	foreach my $keyPage (&KEY_PAGES)
	{
		my $keyUrl = viewUrl( $keyPage ) ;
		print "| <a href=\"$keyUrl\">$keyPage</a>\n" ;
	}
	print <<EOF ;
| </p>
</body>
</html>
EOF
}

sub viewUrl
{
	my $page = shift @_ ;
	return WIKI_BASE . "view.pl?page=" . $page ;
}

sub editUrl
{
	my $page = shift @_ ;
	return WIKI_BASE . "edit.pl?page=" . $page ;
}

sub updateUrl
{
	my $page = shift @_ ;
	return WIKI_BASE . "update.pl?page=" . $page ;
}

sub sourceUrl
{
	my $page = shift @_ ;
	return WIKI_BASE . "source.pl?page=" . $page ;
}

# todo: use relative URLs

sub validatorUrl # todo: use referer-based form, to ease move to relative URLs
{
	my $page = shift @_ ;
	my $viewUrl = viewUrl( $page ) ;
	return "http://validator.w3.org/check?uri=" . $viewUrl ;
}

sub backlinksUrl
{
        my $page = shift @_ ;
        return WIKI_BASE . "backlinks.pl?page=" . $page ;
}

sub changesUrl
{
        return WIKI_BASE . "changes.pl" ;
}

sub viewsUrl
{
        return WIKI_BASE . "views.pl" ;
}

sub escape
{
	my $str = shift @_ ;
	$str =~ s/&/&amp;/g ;
	$str =~ s/</&lt;/g ;
	$str =~ s/>/&gt;/g ;
#	$str =~ s/"/&quot;/g ; # not necessary, and validator chokes
	return $str ;
}

### PAGE LOGIC

sub changes # returns a list of page names, sorted by date
{
	my $days = shift @_ ;
	my $num = shift @_ ;
	# could use opendir, could use glob, could use backticks
	opendir( DIR, WIKI_ROOT ) ;
	my @files = readdir( DIR ) ;
	my @changes = pages() ;
	# todo: more efficient to do this after the sort, binary-chopping to the cutoff and then slicing the list
	if ( $days )
	{
		my $deadline = time() - ($days * 24 * 60 * 60) ;
		@changes = grep {pageMtime( $_ ) ge $deadline} @changes ;
	}
	@changes = sort { pageMtime( $b ) <=> pageMtime( $a ) } @changes ;
	if ( $num )
	{
		my $len = scalar @changes ;
		my $lim = (($num <= $len) ? $num : $len) - 1 ;
		@changes = @changes[0..$lim] ;
	}
	return @changes ;
}

sub views # returns a list of page names, sorted by date
{
	my $days = shift @_ ;
	my $num = shift @_ ;
	# could use opendir, could use glob, could use backticks
	opendir( DIR, WIKI_ROOT ) ;
	my @files = readdir( DIR ) ;
	my @views = pages() ;
	# todo: more efficient to do this after the sort, binary-chopping to the cutoff and then slicing the list
	if ( $days )
	{
		my $deadline = time() - ($days * 24 * 60 * 60) ;
		@views = grep {pageAtime( $_ ) ge $deadline} @views ;
	}
	@views = sort { pageAtime( $b ) <=> pageAtime( $a ) } @views ;
	if ( $num )
	{
		my $len = scalar @views ;
		my $lim = (($num <= $len) ? $num : $len) - 1 ;
		@views = @views[0..$lim] ;
	}
	return @views ;
}

sub backlinks
{
	my $page = shift @_ ;
	my @pages = pages() ;
	my @backlinks = grep { isLink( $_, $page ) } @pages ;
	return @backlinks ;
}

sub isLink
{
	my $from = shift @_ ;
	my $to = shift @_ ;
	my $fh = openPage( $from ) ;
	my $hit = 0 ;
	while (!$hit && (my $line = <$fh>))
	{
		my @atoms = atomise( $line ) ; # todo: it feels a bit icky calling this - atomise is formatting code (?)
		$hit = inList( $to, @atoms ) ;
	}
	close( $fh ) ;
	return $hit ;
}

sub pageExists
{
	my $name = shift @_ ;
	my $file = pageFile( $name ) ;
	return -e $file ;
}

sub pages # list of all page names
{
	opendir( DIR, WIKI_ROOT ) ;
	my @files = readdir( DIR ) ;
	my @pages ;
	foreach my $file (@files)
	{
		if (my $stem = chopExt( $file, WIKI_EXT ))
		{
			push( @pages, validateWikiName( $stem ) ) ;
		}
	}
	closedir( DIR ) ;
	return @pages ;
}

sub pageMtime
{
	my $page = shift @_ ;
	my $file = pageFile( $page ) ;
	return mtime( $file ) ;
}

sub pageAtime
{
	my $page = shift @_ ;
	my $file = pageFile( $page ) ;
	return atime( $file ) ;
}

### UTILITIES

sub chopExt
{
	my $str = shift @_ ;
	my $ext = shift @_ ;
	my $extLen = length( $ext ) ;
	if (substr( $str, -$extLen) eq $ext)
	{
		return substr( $str, 0, -$extLen ) ;
	}
	else
	{
		return undef ;
	}
}

sub inList
{
	my $tgt = shift @_ ;
	my @lst = @_ ;
	foreach my $item (@lst)
	{
		if ($item eq $tgt)
		{
			return 1 ;
		}
	}
	return 0 ;
}

sub normaliseLines
{
	my $text = shift @_ ;
	return join( "\n", split( /\x0D\x0A|\x0A|\x0D/, $text ) ) . "\n" ;
#	$text =~ s/\x0D\x0A|\x0A|\x0D/\n/ ;
		# todo: make this work
	return $text ;
}

sub validateWikiName
{
	my $page = shift @_ ;
	($page = isWikiName( $page )) || die( "bad page name: $page" ) ;
	return $page ;
}

sub isWikiName
{
	my $str = shift @_ ;
	if ($str =~ /^([A-Z][A-Za-z0-9]*[A-Z][A-Za-z0-9]*)$/)
	{
		return $1 ;
	}
	else
	{
		return undef ;
	}
}

sub isUrl
{
	my $str = shift @_ ;
	if ($str =~ /^@{[URL_RE]}$/)
	{
		if (wantarray)
		{
			return ($1, $2) ;
		}
		else
		{
			return "$1:$2" ;
		}
	}
	else
	{
		return undef ;
	}
}

sub parse # todo: replace with CGI.pm or CI::Lite call?
{
	my $queryString = shift @_ ;
	my @pairs = split( "&", $queryString ) ;
	my %query ;
	foreach my $pair (@pairs)
	{
	        (my $name, my $value) = split( "=", $pair ) ;
		defined( $value ) || die( "bad pair: $pair" ) ;
	        $value =~ tr/+/ / ;
	        $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg ;
	        $name =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg ;
	        $value =~ s/\0//g ;
	        $name  =~ s/\0//g ;
	        $query{$name} = $value ;
	}
	return %query ;
}

### FILE HANDLING

sub pageFile
{
	my $page = shift @_ ;
	return WIKI_ROOT . $page . WIKI_EXT ;
}

sub mtime
{
	my $file = shift @_ ;
	my @stat = stat( $file ) ;
	return $stat[9] ;
}

sub atime
{
	my $file = shift @_ ;
	my @stat = stat( $file ) ;
	return $stat[8] ;
}

sub openPage
{
	my $page = shift @_ ;
	my $mode = shift @_ ;
        my $pageFile = pageFile( $page ) ;
	return fhopen( $pageFile, $mode ) ;
}

sub fhopen
{
        my $file = shift @_ ;
	my $mode = shift @_ ;
	if (!$mode)
	{
		$mode = "<" ;
	}
        local *FH ;
        open( FH, $mode.$file ) || die( "unable to open file $file in mode \'$mode\': $!" ) ;
	# todo: do the appropriate locking here
#	use Fcntl ;
#	my $lockMode = ($mode eq ">") ? LOCK_EX : LOCK_SH ;
#	flock( FH, $lockMode ) || die( "unable to lock file $file in mode \'$mode\': $!" ) ;
        return *FH ;
}
